import functools
import typing

from aws_cdk import App, Environment, Stack

from cdk_resources.utils import (
    app_context,
    ALLOWED_ENVIRONMENTS,
    get_environment,
)


__all__ = ["ResourceStack", "register_stacks"]


class ResourceStack(Stack):
    """
    Base Resource Stack
    """

    IMPORT_RESOURCES = None
    RESOURCES = None

    def __init__(self, scope: App, stack_id: str, **kwargs) -> None:
        super().__init__(scope, stack_id, **kwargs)
        # Update Context
        app_context.update(app=scope, current_stack=self)
        if self.is_valid_environment is False:
            raise Exception(
                f"`{get_environment()}` must be a valid environment allowed "
                f"values {ALLOWED_ENVIRONMENTS}"
            )

        # Import resources
        for resources in self.get_import_resources() or []:
            resource_name, Resource, resource_attrs = (
                self.get_resource_name(resources[0]),
                resources[1],
                (resources[2] if len(resources) == 3 else {}),
            )
            setattr(
                self,
                resource_name,
                Resource(
                    scope=self,
                    construct_id=resources[0],
                    **resource_attrs,
                ),
            )

        # Own Resources
        for resources in self.get_resources() or []:
            resource_name, Resource, resource_attrs = (
                self.get_resource_name(resources[0]),
                resources[1],
                (resources[2] if len(resources) == 3 else {}),
            )
            resource = Resource(scope=self, construct_id=resource_name)
            setattr(self, resource_name, resource)

    @classmethod
    def get_dynamic_resources(cls, resource_attr: str = "resources"):
        env = get_environment().lower()
        env = env.replace("-", "_")
        env_resources_attr = getattr(cls, f"{env.upper()}_{resource_attr.upper()}", None)
        env_resources_method = getattr(cls, f"get_{env}_{resource_attr.lower()}", None)
        if callable(env_resources_method):
            return env_resources_method()
        elif type(env_resources_attr) in [list, tuple]:
            return env_resources_attr
        return getattr(cls, resource_attr.upper())

    @classmethod
    def get_import_resources(cls):
        return cls.get_dynamic_resources("import_resources")

    @classmethod
    def get_resources(cls):
        return cls.get_dynamic_resources("resources")

    @staticmethod
    def get_resource_name(value: typing.Union[str, typing.Callable]) -> str:
        return value() if hasattr(value, "__call__") else value

    @property
    @functools.lru_cache(maxsize=None)
    def is_valid_environment(self) -> bool:
        if len(ALLOWED_ENVIRONMENTS) == 0:
            return True
        environment = get_environment()
        return environment is not None and environment in ALLOWED_ENVIRONMENTS


def register_stacks(app: App, aws_env: Environment, stacks: list) -> None:
    # Create Stacks
    for stack in stacks:
        stack_id, stack_class, stack_kwargs = (
            stack[0],
            stack[1],
            (stack[2] if len(stack) == 3 else {}),
        )
        stack_class(app, stack_id, env=aws_env, **stack_kwargs)



from aws_cdk import aws_ec2

from cdk_resources.resources.base import Resource
from cdk_resources.utils import combine_configurations


__all__ = ("BaseSecurityGroup",)


class BaseSecurityGroup(Resource[aws_ec2.SecurityGroup]):
    # Others
    sg_ingress_rules_props = None

    @classmethod
    def post_create(cls, construct: aws_ec2.SecurityGroup) -> None:
        sg_ingress_rules_props = combine_configurations(
            cls.sg_ingress_rules_props or {}
        )
        for rules in sg_ingress_rules_props["ingress_rules"]:
            construct.add_ingress_rule(**rules)



import os
import typing
import functools

from aws_cdk import App


# Constants
ALLOWED_ENVIRONMENTS = os.environ.get("ALLOWED_ENVIRONMENTS", [])
COMMON_ENVIRONMENT_KEY = os.environ.get("COMMON_ENVIRONMENT_KEY", "default")
ENVIRONMENT_CONTEXT_KEY = os.environ.get(
    "ENVIRONMENT_CONTEXT_KEY", "environment"
)

# Shared var
app_context = {}


aws_account = os.environ.get("CDK_DEFAULT_ACCOUNT")
aws_region = os.environ.get("CDK_DEFAULT_REGION")


def combine_configurations(
    config: typing.Any,
    key: typing.Optional[str] = None,
    exclude_keys: typing.Optional[list] = None,
) -> typing.Any:
    exclude_keys = exclude_keys or []

    def filter_config(conf: dict) -> dict:
        if aws_account and aws_account in conf:
            return filter_config(conf[aws_account])
        elif aws_region and aws_region in conf:
            return filter_config(conf[aws_region])

        filtered_config = {
            k: v for k, v in conf.items() if k not in exclude_keys
        }
        return filtered_config if key is None else filtered_config.get(key, {})

    if isinstance(config, dict) is False:
        return config

    # Environments
    environment = get_environment()
    if environment is None:
        # Not environment app
        return filter_config(config.get(COMMON_ENVIRONMENT_KEY, config))

    environments = ALLOWED_ENVIRONMENTS + [COMMON_ENVIRONMENT_KEY]
    env_config = (
        {k: v for k, v in config.items() if k in environments}
        if len(ALLOWED_ENVIRONMENTS) > 0
        else config
    )
    if len(env_config) == 0:
        # Not environment app
        return filter_config(config)

    combined_config = dict(
        env_config.get(COMMON_ENVIRONMENT_KEY, {}),
        **env_config.get(environment, {}),
    )
    return filter_config(combined_config)


def get_context_variable(key: str) -> typing.Any:
    context_variable = app_context["app"].node.try_get_context(key)
    context_variable = combine_configurations(context_variable)
    return context_variable


@functools.lru_cache(maxsize=None)
def get_environment(app: typing.Optional[App] = None) -> typing.Optional[str]:
    environment = (app or app_context["app"]).node.try_get_context(
        ENVIRONMENT_CONTEXT_KEY
    ) or os.getenv(ENVIRONMENT_CONTEXT_KEY.upper())
    app_context[ENVIRONMENT_CONTEXT_KEY] = environment
    return environment


def env(**kwargs):
    """
    """
    environment = get_environment()

    def filter_env(conf: dict):
        if not isinstance(conf, dict):
            return conf
        if environment in conf:
            return filter_env(conf.get(environment))
        if aws_account and aws_account in conf:
            return filter_env(conf.get(aws_account))
        if aws_region and aws_region in conf:
            return filter_env(conf.get(aws_region))
        if COMMON_ENVIRONMENT_KEY in conf:
            return filter_env(conf.get(COMMON_ENVIRONMENT_KEY))
        return conf

    return filter_env(kwargs)


@functools.lru_cache(maxsize=None)
def context_var(var: str, default: typing.Any = None, app: typing.Optional[App] = None, required: bool = False) -> typing.Optional[str]:
    app = app or app_context["app"]
    value = app.node.try_get_context(var) or os.getenv(var)  or default
    if value is None and required:
        raise Exception(f"{var} context variable is required")
    return value



import abc
import functools
import inspect
import typing

from aws_cdk import App, Stack, Tags
from constructs import Construct

from cdk_resources.utils import (
    combine_configurations,
    app_context,
    get_environment,
)


__all__ = ["Resource", "ResourceType", "ResourceTagMixin"]


ResourceType = typing.TypeVar("ResourceType")


class Resource(typing.Generic[ResourceType], abc.ABC):
    """ """

    # Main Construct (Required)
    @classmethod
    @abc.abstractmethod
    def construct(cls, scope: Stack, id: str, **kwargs) -> ResourceType:
        pass

    # Environment Props
    environment_props = None

    def __new__(
        cls,
        scope: typing.Optional[Stack] = None,
        construct_id: typing.Optional[str] = None,
        force_lookup: bool = False,
        **kwargs,
    ) -> ResourceType:
        if hasattr(cls, "construct_cache") and force_lookup is False:
            return cls
        scope = scope or app_context["current_stack"]
        assert isinstance(scope, Stack), "Scope is required"
        assert isinstance(construct_id, str), "construct_id is required"
        construct = cls.create(scope, construct_id, **kwargs)
        setattr(cls, "construct_cache", construct)
        return cls

    @classmethod
    def get(cls) -> ResourceType:
        construct_cache = getattr(cls, "construct_cache", None)
        assert construct_cache is not None, f"{cls} has not being initializated"
        return construct_cache

    @classmethod
    def create(cls, scope: Stack, construct_id: str, **kwargs) -> ResourceType:
        kwargs = dict(scope=scope, id=construct_id, **kwargs)
        # Environment Props
        environment_props = cls.get_environment_props()
        cls.pre_create(environment_props)
        kwargs.update(environment_props)
        # Create
        environment_construct = getattr(
            cls, f"construct_{cls.environment}", None
        )
        construct_method = (
            environment_construct
            if callable(environment_construct)
            else cls.construct
        )
        construct = construct_method(**kwargs)
        # Post Create
        cls.post_create(construct)
        return construct

    @classmethod
    def get_environment_props(cls) -> dict:
        environment_props = cls.environment_props or {}
        if callable(environment_props):
            environment_props = environment_props()
        environment_props = combine_configurations(environment_props)
        return environment_props

    @classmethod
    def pre_create(cls, environment_props: dict) -> None:
        pass

    @classmethod
    def post_create(cls, construct: ResourceType) -> None:
        pass

    @classmethod
    @property
    def environment(cls):
        return get_environment()


class ResourceTagMixin:
    construct_tags = None

    @classmethod
    def post_create(cls, construct: ResourceType) -> None:
        super().post_create(construct)
        for key, value in cls.construct_tags or []:
            value = value if isinstance(value, str) else value()
            Tags.of(construct).add(key, value)



import os
import setuptools


base_dir = os.path.abspath(os.path.dirname(__file__))

# Readme
with open("README.md") as fp:
    readme = fp.read()

# About
version = {}
with open(
    os.path.join(base_dir, "cdk_resources", "__version__.py"), encoding="utf-8"
) as f:
    exec(f.read(), version)

setuptools.setup(
    name=version["__name__"],
    version=version["__version__"],
    description=version["__description__"],
    long_description=readme,
    long_description_content_type="text/markdown",
    author=version["__author__"],
    packages=setuptools.find_packages(
        exclude=["tests*", "samples*", "mkdocs*"]
    ),
    # package_dir={"": "cdk_resources"},
    package_data={"": ["LICENSE"]},
    install_requires=["aws-cdk-lib>=2.43.1", "boto3>=1.20.23"],
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: AWS CDK :: 1",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
)
